#ifndef RPC_EXCEPTION_SERIALIZATION_H
#define RPC_EXCEPTION_SERIALIZATION_H

#include "RpcTypelist.h"
#include "RpcTypes.h"
#include "RpcIncorrectTypeId.h"

namespace rpc
{
    template<class Exception, int id>
    struct exception_serialize_traits
    {
        // pack by value
        template<class ArchiveType>
        static void serialize(ArchiveType& archive, const Exception& object)
        {
            int* typeId = (int*)archive.TailAllocate(sizeof(int));
            *typeId = id;

            rpc_size_t* msgSize = (rpc_size_t*)archive.TailAllocate(sizeof(rpc_size_t));
            size_t length = strlen(object.what());
            *msgSize = length;

            char* message = (char*)archive.TailAllocate(length);
            memcpy(message, object.what(), length);
        }

        // unpack by ptr
        template<class ArchiveType>
        static void deserialize(ArchiveType& archive, Exception* object)
        {
            int* typeId = (int*)archive.FetchRange(sizeof(int));
            if (*typeId != id)
                throw rpc::IncorrectTypeId(id, *typeId);

            rpc_size_t* msgSize = (rpc_size_t*)archive.FetchRange(sizeof(rpc_size_t));
            size_t length = (size_t)*msgSize;

            const char* message = (const char*)archive.FetchRange(length);
            std::string temp(message, length);
            *object = Exception(temp.c_str());
        }
    };

    template<> struct serialize_traits<std::exception>:public exception_serialize_traits<std::exception, rid_std_exception>{};
    template<> struct serialize_traits<std::bad_alloc>:public exception_serialize_traits<std::bad_alloc, rid_std_bad_alloc>{};
    template<> struct serialize_traits<std::bad_cast>:public exception_serialize_traits<std::bad_cast, rid_std_bad_cast>{};
    template<> struct serialize_traits<std::bad_exception>:public exception_serialize_traits<std::bad_exception, rid_std_bad_exception>{};
    template<> struct serialize_traits<std::bad_typeid>:public exception_serialize_traits<std::bad_typeid, rid_std_bad_typeid>{};
    template<> struct serialize_traits<std::runtime_error>:public exception_serialize_traits<std::runtime_error, rid_std_runtime_error>{};
    template<> struct serialize_traits<std::range_error>:public exception_serialize_traits<std::range_error, rid_std_range_error>{};
    template<> struct serialize_traits<std::overflow_error>:public exception_serialize_traits<std::overflow_error, rid_std_overflow_error>{};
    template<> struct serialize_traits<std::underflow_error>:public exception_serialize_traits<std::underflow_error, rid_std_underflow_error>{};
    template<> struct serialize_traits<std::logic_error>:public exception_serialize_traits<std::logic_error, rid_std_runtime_error>{};
    template<> struct serialize_traits<std::domain_error>:public exception_serialize_traits<std::domain_error, rid_std_runtime_error>{};
    template<> struct serialize_traits<std::invalid_argument>:public exception_serialize_traits<std::invalid_argument, rid_std_runtime_error>{};
    template<> struct serialize_traits<std::length_error>:public exception_serialize_traits<std::length_error, rid_std_runtime_error>{};
    template<> struct serialize_traits<std::out_of_range>:public exception_serialize_traits<std::out_of_range, rid_std_runtime_error>{};

    //////////////////////////////////////////////////////////////////////////

    template<class ExTypeList, unsigned int inx>
    struct UnpackAndThrowImpl
    {
        template<class ArchiveType>
        inline static void Do(ArchiveType& archive, unsigned int index)
        {
            if (index == inx)
            {
                typename rpc::TL::TypeAt<ExTypeList, inx>::Result exd("");
                rpc::Unpack(archive, &exd);
                throw exd;
            }

            UnpackAndThrowImpl<ExTypeList, inx - 1>::Do(archive, index);
        }
    };

    template<class ExTypeList>
    struct UnpackAndThrowImpl<ExTypeList, -1>
    {
        template<class ArchiveType>
        inline static void Do(ArchiveType&, unsigned int)
        {
            return;
        }
    };

    //////////////////////////////////////////////////////////////////////////

    template<class ExTypeList, unsigned int inx>
    struct DispatchImpl
    {
        template<class ArchiveType, class Command>
        inline static void Do(ArchiveType& archive, Command& command)
        {
            try
            {
                DispatchImpl<ExTypeList, inx - 1>::Do(archive, command);
            }
            catch (const typename TL::TypeAt<ExTypeList, inx>::Result &ex)
            {
                archive.UnTail(archive.GetUsedSize());
                rpc::Pack(archive, true); // exception flag
                rpc::Pack(archive, inx);
                rpc::Pack(archive, ex);
            }
        }
    };

    template<class ExTypeList>
    struct DispatchImpl<ExTypeList, -1>
    {
        template<class ArchiveType, class Command>
        inline static void Do(ArchiveType&, Command& command)
        {
            return command();
        }
    };

    //////////////////////////////////////////////////////////////////////////

    template <class ExTypeList>
    struct ExceptionPacker
    {
        template<class ArchiveType>
        inline static void UnpackAndThrow(ArchiveType& archive)
        {
            typedef typename TL::DerivedToFront<ExTypeList>::Result ExTypeList_Ordered;

            unsigned int x;
            rpc::Unpack(archive, &x);

            UnpackAndThrowImpl<ExTypeList_Ordered, TL::Length<ExTypeList_Ordered>::value-1 >::Do(archive, x);
        }

        template<class ArchiveType, class ExceptionType>
        inline static void Pack(ArchiveType& archive, const ExceptionType& exs)
        {
            typedef typename TL::DerivedToFront<ExTypeList>::Result ExTypeList_Ordered;

            rpc::Pack(archive, (unsigned int)TL::IndexOf<ExTypeList_Ordered, ExceptionType>::value);
            rpc::Pack(archive, exs);
        }

        template<class ArchiveType ,class Command>
        inline static void Dispatch(ArchiveType& archive, Command& command)
        {
            typedef typename TL::DerivedToFront<ExTypeList>::Result ExTypeList_Ordered;

            DispatchImpl<ExTypeList_Ordered, TL::Length<ExTypeList_Ordered>::value - 1 >::Do(archive, command);
        }

    };

}

#endif
