#ifndef RPC_SERIALIZATION_H
#define RPC_SERIALIZATION_H

#include "stdexcept"
#include "typeinfo"

#include "RpcTypes.h"
#include "RpcIncorrectTypeId.h"

#include "StlSerialization.h"
#include "RpcExceptionSerialization.h"

//////////////////////////////////////////////////////////////////////////

namespace rpc
{

    //////////////////////////////////////////////////////////////////////////

    template<class PODType>
    struct serialize_traits_pod_impl
    {
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const PODType & object)
        {
            int * pTypeId = (int * )archive.TailAllocate(sizeof(int));
            *pTypeId = (int)resource_traits<PODType>::ResourceId;
            PODType * pData = (PODType * )archive.TailAllocate(sizeof(object));
            memcpy(pData, &object, sizeof(object));
        }

        template<class ArchiveType>
        static void deserialize(ArchiveType & archive, PODType  * pObject)
        {
            int iTypeId = (int)resource_traits<PODType>::ResourceId;
            int * pTypeId = (int * )archive.FetchRange(sizeof(int) );
            if (*pTypeId != iTypeId)
                throw rpc::IncorrectTypeId(iTypeId, *pTypeId);

            PODType  * pSourceObject = (PODType  * )archive.FetchRange(sizeof(PODType));
            memcpy(pObject, pSourceObject, sizeof(PODType));
        }
    };

    template<> struct serialize_traits<bool>:public serialize_traits_pod_impl<bool>{};
    template<> struct serialize_traits<char>:public serialize_traits_pod_impl<char>{};
    template<> struct serialize_traits<unsigned char>:public serialize_traits_pod_impl<unsigned char>{};
    template<> struct serialize_traits<short>:public serialize_traits_pod_impl<short>{};
    template<> struct serialize_traits<unsigned short>:public serialize_traits_pod_impl<unsigned short>{};
    template<> struct serialize_traits<int>:public serialize_traits_pod_impl<int>{};
    template<> struct serialize_traits<unsigned int>:public serialize_traits_pod_impl<unsigned int>{};
    template<> struct serialize_traits<long>:public serialize_traits_pod_impl<long>{};
    template<> struct serialize_traits<unsigned long>:public serialize_traits_pod_impl<unsigned long>{};
    template<> struct serialize_traits<long long>:public serialize_traits_pod_impl<long long>{};
    template<> struct serialize_traits<unsigned long long>:public serialize_traits_pod_impl<unsigned long>{};
    template<> struct serialize_traits<wchar_t>:public serialize_traits_pod_impl<wchar_t>{};
    template<> struct serialize_traits<float>:public serialize_traits_pod_impl<double>{};
    template<> struct serialize_traits<double>:public serialize_traits_pod_impl<double>{};

    template<class pointer>
    struct serialize_traits_pointer_impl
    {
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, pointer ptr) 
        {
            int * pTypeId = (int* )archive.TailAllocate(sizeof(int));
            *pTypeId = (int)resource_traits<pointer>::ResourceId;
            pointer* pData = (pointer* )archive.TailAllocate(sizeof(ptr));
            *pData = ptr;
        }

        template<class ArchiveType>
        static void deserialize(ArchiveType & archive, pointer* pptr)
        {
            int iTypeId = (int)resource_traits<pointer>::ResourceId;
            int* pTypeId = (int* )archive.FetchRange(sizeof(int));
            if (*pTypeId != iTypeId)
                throw rpc::IncorrectTypeId(iTypeId, *pTypeId);

            pointer* pSourcePtr = (pointer* )archive.FetchRange(sizeof(pointer));
            *pptr = *pSourcePtr;
        }
    };

    template<> struct serialize_traits<const void*>:public serialize_traits_pointer_impl<const void*>{};
    template<> struct serialize_traits<void*>:public serialize_traits_pointer_impl<void*>{};


    //////////////////////////////////////////////////////////////////////////
    // PACK & UNPACK

    template<class ArchiveType, class ObjectType>
    void Pack(ArchiveType & archive, const ObjectType& object)
    {
        serialize_traits<ObjectType>::serialize(archive, object);
    }

    template<class ArchiveType, class ObjectType>
    void Unpack(ArchiveType & archive, ObjectType* pObject)
    {
        serialize_traits<ObjectType>::deserialize(archive, pObject);
    }

    //////////////////////////////////////////////////////////////////////////
    // ARRAY

    template<class ObjectType>
    struct pack_helper
    {
        template<class ArchiveType>
        static void PackEmptyArray(ArchiveType & archive)
        {
            int  * pTypeId = (int  *)archive.TailAllocate(sizeof (int));
            *pTypeId = rid_sequence;

            int  * pNumberOfDimensions = (int  *)archive.TailAllocate(sizeof (int));
            *pNumberOfDimensions = 1;

            rpc_size_t  * pSize = (rpc_size_t  *)archive.TailAllocate(sizeof (rpc_size_t));
            *pSize = 0;
        }
    };

    template<class ArchiveType, class ObjectType>
    void PackArray(ArchiveType & archive, ObjectType * pObject, rpc_size_t size)
    {
        int  * pTypeId = (int  *)archive.TailAllocate(sizeof (int));
        *pTypeId = rid_sequence;

        int  * pNumberOfDimensions = (int  *)archive.TailAllocate(sizeof (int));
        *pNumberOfDimensions = 1;

        rpc_size_t  * pSize = (rpc_size_t  *)archive.TailAllocate(sizeof (rpc_size_t));
        *pSize = size;

        for(rpc_size_t i = 0; i<size; ++i)
            Pack(archive, pObject[i]);
    }

    template<class ArchiveType>
    void PackArray(ArchiveType & archive, void * pObject, rpc_size_t size)
    {
        PackArray(archive, (char*)pObject, size);
    }

    template<class ArchiveType, class ObjectType>
    void UnpackArray(ArchiveType & archive, ObjectType * pObject, rpc_size_t size_to_check)
    {
        int  * pTypeId = (int  *)archive.FetchRange(sizeof (int));
        if (*pTypeId != rid_sequence)
            throw rpc::IncorrectTypeId(rid_sequence, *pTypeId);

        int  * pNumberOfDimensions = (int  *)archive.FetchRange(sizeof (int));
        if (*pNumberOfDimensions != 1)
            throw std::runtime_error("Array deserialization failed - incorrect dimension");

        rpc_size_t * pDim1 = (rpc_size_t  *)archive.FetchRange(sizeof (rpc_size_t));
        if (*pDim1 > size_to_check)
            throw std::runtime_error("Array deserialization failed - incorrect size returned");

        for(rpc_size_t i = 0; i<*pDim1; ++i)
            Unpack(archive, &(pObject[i]));
    }
} //rpc

#endif