#ifndef RPC_TYPELIST_H
#define RPC_TYPELIST_H

#include "RpcTypelistUtils.h"

namespace rpc
{
    class NullType {};

    ////////////////////////////////////////////////////////////////////////////////
    // class template Typelist
    // The building block of typelists of any length
    // Defines nested types:
    //     Head (first element, a non-typelist type by convention)
    //     Tail (second element, can be another typelist)
    ////////////////////////////////////////////////////////////////////////////////

    template <class T, class U>
    struct Typelist
    {
        typedef T Head;
        typedef U Tail;
    };

    namespace TL
    {
        ////////////////////////////////////////////////////////////////////////////////
        // class template MakeTypelist
        // Takes a number of arguments equal to its numeric suffix
        // The arguments are type names.
        // MakeTypelist<T1, T2, ...>::Result
        // returns a typelist that is of T1, T2, ...
        ////////////////////////////////////////////////////////////////////////////////

        template
            <
            typename T1  = NullType, typename T2  = NullType, typename T3  = NullType,
            typename T4  = NullType, typename T5  = NullType, typename T6  = NullType,
            typename T7  = NullType, typename T8  = NullType, typename T9  = NullType,
            typename T10 = NullType, typename T11 = NullType, typename T12 = NullType,
            typename T13 = NullType, typename T14 = NullType, typename T15 = NullType,
            typename T16 = NullType, typename T17 = NullType, typename T18 = NullType
            >
        struct MakeTypelist
        {
        private:
            typedef typename MakeTypelist
                <
                T2 , T3 , T4 ,
                T5 , T6 , T7 ,
                T8 , T9 , T10,
                T11, T12, T13,
                T14, T15, T16,
                T17, T18
                >
                ::Result TailResult;

        public:
            typedef Typelist<T1, TailResult> Result;
        };

        template<>
        struct MakeTypelist<>
        {
            typedef NullType Result;
        };

        ////////////////////////////////////////////////////////////////////////////////
        // class template Append
        // Appends a type or a typelist to another
        // Invocation (TList is a typelist and T is either a type or a typelist):
        // Append<TList, T>::Result
        // returns a typelist that is TList followed by T and NullType-terminated
        ////////////////////////////////////////////////////////////////////////////////

        template <class TList, class T> struct Append;

        template <> struct Append<NullType, NullType>
        {
            typedef NullType Result;
        };

        template <class T> struct Append<NullType, T>
        {
            typedef Typelist<T,NullType> Result;
        };

        template <class Head, class Tail>
        struct Append<NullType, Typelist<Head, Tail> >
        {
            typedef Typelist<Head, Tail> Result;
        };

        template <class Head, class Tail, class T>
        struct Append<Typelist<Head, Tail>, T>
        {
            typedef Typelist<Head,
                typename Append<Tail, T>::Result>
                Result;
        };

        ////////////////////////////////////////////////////////////////////////////////
        // class template TypeAt
        // Finds the type at a given index in a typelist
        // Invocation (TList is a typelist and index is a compile-time integral
        //     constant):
        // TypeAt<TList, index>::Result
        // returns the type in position 'index' in TList
        // If you pass an out-of-bounds index, the result is a compile-time error
        ////////////////////////////////////////////////////////////////////////////////

        template <class TList, unsigned int index> struct TypeAt;

        template <class Head, class Tail>
        struct TypeAt<Typelist<Head, Tail>, 0>
        {
            typedef Head Result;
        };

        template <class Head, class Tail, unsigned int i>
        struct TypeAt<Typelist<Head, Tail>, i>
        {
            typedef typename TypeAt<Tail, i - 1>::Result Result;
        };

        ////////////////////////////////////////////////////////////////////////////////
        // class template IndexOf
        // Finds the index of a type in a typelist
        // Invocation (TList is a typelist and T is a type):
        // IndexOf<TList, T>::value
        // returns the position of T in TList, or NullType if T is not found in TList
        ////////////////////////////////////////////////////////////////////////////////

        template <class TList, class T> struct IndexOf;

        template <class T>
        struct IndexOf<NullType, T>
        {
            enum { value = -1 };
        };

        template <class T, class Tail>
        struct IndexOf<Typelist<T, Tail>, T>
        {
            enum { value = 0 };
        };

        template <class Head, class Tail, class T>
        struct IndexOf<Typelist<Head, Tail>, T>
        {
        private:
            enum { temp = IndexOf<Tail, T>::value };
        public:
            enum { value = (temp == -1 ? -1 : 1 + temp) };
        };

        ////////////////////////////////////////////////////////////////////////////////
        // class template Length
        // Computes the length of a typelist
        // Invocation (TList is a typelist):
        // Length<TList>::value
        // returns a compile-time constant containing the length of TList, not counting
        //     the end terminator (which by convention is NullType)
        ////////////////////////////////////////////////////////////////////////////////

        template <class TList> struct Length;
        template <> struct Length<NullType>
        {
            enum { value = 0 };
        };

        template <class T, class U>
        struct Length< Typelist<T, U> >
        {
            enum { value = 1 + Length<U>::value };
        };

        ////////////////////////////////////////////////////////////////////////////////
        // class template Replace
        // Replaces the first occurence of a type in a typelist, with another type
        // Invocation (TList is a typelist, T, U are types):
        // Replace<TList, T, U>::Result
        // returns a typelist in which the first occurence of T is replaced with U
        ////////////////////////////////////////////////////////////////////////////////

        template <class TList, class T, class U> struct Replace;

        template <class T, class U>
        struct Replace<NullType, T, U>
        {
            typedef NullType Result;
        };

        template <class T, class Tail, class U>
        struct Replace<Typelist<T, Tail>, T, U>
        {
            typedef Typelist<U, Tail> Result;
        };

        template <class Head, class Tail, class T, class U>
        struct Replace<Typelist<Head, Tail>, T, U>
        {
            typedef Typelist<Head,
                typename Replace<Tail, T, U>::Result>
                Result;
        };

        ////////////////////////////////////////////////////////////////////////////////
        // class template Select
        // Selects one of two types based upon a boolean constant
        // Invocation: Select<flag, T, U>::Result
        // where:
        // flag is a compile-time boolean constant
        // T and U are types
        // Result evaluates to T if flag is true, and to U otherwise.
        ////////////////////////////////////////////////////////////////////////////////

        template <bool flag, typename T, typename U>
        struct Select
        {
            typedef T Result;
        };
        template <typename T, typename U>
        struct Select<false, T, U>
        {
            typedef U Result;
        };

        ////////////////////////////////////////////////////////////////////////////////
        // class template MostDerived
        // Finds the type in a typelist that is the most derived from a given type
        // Invocation (TList is a typelist, T is a type):
        // MostDerived<TList, T>::Result
        // returns the type in TList that's the most derived from T
        ////////////////////////////////////////////////////////////////////////////////

        template <class TList, class T> struct MostDerived;

        template <class T>
        struct MostDerived<NullType, T>
        {
            typedef T Result;
        };

        template <class Head, class Tail, class T>
        struct MostDerived<Typelist<Head, Tail>, T>
        {
        private:
            typedef typename MostDerived<Tail, T>::Result Candidate;
        public:
            typedef typename Select<
                SuperSubclass<Candidate,Head>::value,
                Head, Candidate>::Result Result;
        };

        ////////////////////////////////////////////////////////////////////////////////
        // class template DerivedToFront
        // Arranges the types in a typelist so that the most derived types appear first
        // Invocation (TList is a typelist):
        // DerivedToFront<TList>::Result
        // returns the reordered TList
        ////////////////////////////////////////////////////////////////////////////////

        template <class TList> struct DerivedToFront;

        template <>
        struct DerivedToFront<NullType>
        {
            typedef NullType Result;
        };

        template <class Head, class Tail>
        struct DerivedToFront< Typelist<Head, Tail> >
        {
        private:
            typedef typename MostDerived<Tail, Head>::Result
                TheMostDerived;
            typedef typename Replace<Tail,
                TheMostDerived, Head>::Result Temp;
            typedef typename DerivedToFront<Temp>::Result L;
        public:
            typedef Typelist<TheMostDerived, L> Result;
        };
    }
}
#endif
