#ifndef RPC_TYPELIST_UTILS_H
#define RPC_TYPELIST_UTILS_H

namespace rpc
{
    ////////////////////////////////////////////////////////////////////////////////
    // Helper types Small and Big - guarantee that sizeof(Small) < sizeof(Big)
    ////////////////////////////////////////////////////////////////////////////////

    namespace Private
    {
        template <class T, class U>
        struct ConversionHelper
        {
            typedef char Small;
            struct Big { char dummy[2]; };
            static Big   Test(...);
            static Small Test(U);
            static T MakeT();
        };
    }

    ////////////////////////////////////////////////////////////////////////////////
    // class template Conversion
    // Figures out the conversion relationships between two types
    // Invocations (T and U are types):
    // a) Conversion<T, U>::exists
    // returns (at compile time) true if there is an implicit conversion from T
    // to U (example: Derived to Base)
    // b) Conversion<T, U>::exists2Way
    // returns (at compile time) true if there are both conversions from T
    // to U and from U to T (example: int to char and back)
    // c) Conversion<T, U>::sameType
    // returns (at compile time) true if T and U represent the same type
    //
    // Caveat: might not work if T and U are in a private inheritance hierarchy.
    ////////////////////////////////////////////////////////////////////////////////

    template <class T, class U>
    struct Conversion
    {
        typedef Private::ConversionHelper<T, U> H;
        enum { exists = sizeof(typename H::Small) == sizeof((H::Test(H::MakeT()))) };
        enum { exists2Way = exists && Conversion<U, T>::exists };
        enum { sameType = false };
    };

    template <class T>
    struct Conversion<T, T>    
    {
        enum { exists = 1, exists2Way = 1, sameType = 1 };
    };

    template <class T>
    struct Conversion<void, T>    
    {
        enum { exists = 0, exists2Way = 0, sameType = 0 };
    };

    template <class T>
    struct Conversion<T, void>    
    {
        enum { exists = 0, exists2Way = 0, sameType = 0 };
    };

    template <>
    struct Conversion<void, void>    
    {
    public:
        enum { exists = 1, exists2Way = 1, sameType = 1 };
    };

    ////////////////////////////////////////////////////////////////////////////////
    // class template SuperSubclass
    // Invocation: SuperSubclass<B, D>::value where B and D are types. 
    // Returns true if B is a public base of D, or if B and D are aliases of the 
    // same type.
    //
    // Caveat: might not work if T and U are in a private inheritance hierarchy.
    ////////////////////////////////////////////////////////////////////////////////

    template <class T, class U>
    struct SuperSubclass
    {
        enum { value = (Conversion<const volatile U*, const volatile T*>::exists &&
            !Conversion<const volatile T*, const volatile void*>::sameType) };

        // Dummy enum to make sure that both classes are fully defined.
        enum{ dontUseWithIncompleteTypes = ( sizeof (T) == sizeof (U) ) };
    };

    template <>
    struct SuperSubclass<void, void> 
    {
        enum { value = false };
    };

    template <class U>
    struct SuperSubclass<void, U> 
    {
        enum { value = (Conversion<const volatile U*, const volatile void*>::exists &&
            !Conversion<const volatile void*, const volatile void*>::sameType) };

        // Dummy enum to make sure that both classes are fully defined.
        enum{ dontUseWithIncompleteTypes = ( 0 == sizeof (U) ) };
    };

    template <class T>
    struct SuperSubclass<T, void> 
    {
        enum { value = (Conversion<const volatile void*, const volatile T*>::exists &&
            !Conversion<const volatile T*, const volatile void*>::sameType) };

        // Dummy enum to make sure that both classes are fully defined.
        enum{ dontUseWithIncompleteTypes = ( sizeof (T) == 0 ) };
    };
}

#endif