#ifndef RPC_STL_SERIALIZATION_H
#define RPC_STL_SERIALIZATION_H

#include "vector"
#include "list"
#include "map"
#include "string"
#include "cstring"

#include "RpcTypes.h"
#include "RpcIncorrectTypeId.h"

namespace rpc
{

    /// BASIC_STRING

    template<class CharType>
    struct serialize_traits<std::basic_string<CharType> >
    {
        typedef std::basic_string<CharType> UsedString;
        // pack by value 
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const UsedString & object)
        {
            int  * pTypeId = (int  *)archive.TailAllocate(sizeof (int));
            *pTypeId = rid_basic_string;
            int  * pCharSize = (int  *)archive.TailAllocate(sizeof (int));
            *pCharSize = sizeof(CharType);
            rpc_size_t  * pVectorSize = (rpc_size_t  *)archive.TailAllocate(sizeof (rpc_size_t));
            *pVectorSize = object.size();

            char  * pData = (char  *)archive.TailAllocate((size_t)(*pVectorSize * sizeof(CharType)));
            memcpy(pData, &object[0], (size_t)(*pVectorSize * sizeof(CharType)));
        }
        // pack by pointer
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const UsedString * pObject)
        {
            serialize(archive, *pObject);
        }

        // unpack by ptr 
        template<class ArchiveType>
        static void deserialize(ArchiveType & archive, UsedString * ppObject)
        {
            int  *pTypeId = (int  *)archive.FetchRange(sizeof (int));
            if (*pTypeId != rid_basic_string)
                throw rpc::IncorrectTypeId(rid_basic_string, *pTypeId);

            int  * pCharSize = (int  *)archive.FetchRange(sizeof (int));
            if (*pCharSize != sizeof(CharType))
                throw std::runtime_error("basic_string deserialization failed - incorrect char size");

            rpc_size_t  * pVectorSize = (rpc_size_t  *)archive.FetchRange(sizeof (rpc_size_t));
            const char* pData= (const char *)archive.FetchRange((size_t)(*pVectorSize * sizeof(CharType)));

            ppObject->resize((size_t)*pVectorSize);

            memcpy(&(*ppObject)[0], pData, (size_t)(*pVectorSize * sizeof(CharType)));
        }
    };


    /// VECTOR

    template<class ValueType, template<class T> class Allocator>
    struct serialize_traits<std::vector<ValueType, Allocator> >
    {
        // pack by value 
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const std::vector<ValueType> & object)
        {
            if (object.empty())
                pack_helper<ValueType>::PackEmptyArray(archive);
            else
                PackArray(archive, &object[0], object.size());
        }
        // pack by pointer
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const std::vector<ValueType> * pObject)
        {
            serialize(archive, *pObject);
        }

        // unpack by ptr 
        template<class ArchiveType>
        static void deserialize(ArchiveType & archive, std::vector<ValueType> * ppObject)
        {
            std::vector<ValueType> vec;

            int  * pTypeId = (int  *)archive.FetchRange(sizeof (int));
            if (*pTypeId != rid_sequence)
                throw rpc::IncorrectTypeId(rid_sequence, *pTypeId);

            int  * pNumberOfDimensions = (int  *)archive.FetchRange(sizeof (int));
            if (*pNumberOfDimensions != 1)
                throw std::runtime_error("Array deserialization failed - incorrect dimension");

            rpc_size_t * pDim1 = (rpc_size_t  *)archive.FetchRange(sizeof (rpc_size_t));
            vec.resize((size_t)*pDim1);

            if (vec.empty())
            {
                ppObject->clear();
                return;
            }

            ValueType * pObject = &vec[0];
            for(rpc_size_t i = 0; i<*pDim1; ++i)
                serialize_traits<ValueType>::deserialize(archive, &(pObject[i]));  
            ppObject->assign(vec.begin(), vec.end());
        }
    };

    // VECTOR < CHAR >
    template<template<class T> class Allocator>
    struct serialize_traits<std::vector<char, Allocator> >
    {
        // pack by value 
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const std::vector<char, Allocator> & object)
        {
            int  * pTypeId = (int  *)archive.TailAllocate(sizeof (int));
            *pTypeId = rid_vector_char;
            rpc_size_t  * pVectorSize = (rpc_size_t  *)archive.TailAllocate(sizeof (rpc_size_t));
            *pVectorSize = object.size();

            char  * pData = (char  *)archive.TailAllocate(object.size());

            if (*pVectorSize)
                ::memcpy(pData, &object[0], object.size());
        }
        // pack by pointer
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const std::vector<char, Allocator> * pObject)
        {
            serialize(archive, *pObject);
        }

        // unpack by ptr 
        template<class ArchiveType>
        static void deserialize(ArchiveType & archive, std::vector<char, Allocator> * ppObject)
        {
            int  * pTypeId = (int  *)archive.FetchRange(sizeof (int));
            if (*pTypeId != rid_vector_char)
                throw rpc::IncorrectTypeId(rid_vector_char, *pTypeId);

            rpc_size_t  * pVectorSize = (rpc_size_t  *)archive.FetchRange(sizeof (rpc_size_t));
            const char* pData= (const char *)archive.FetchRange((size_t)*pVectorSize);

            ppObject->resize((size_t)*pVectorSize);

            if (*pVectorSize)
                memcpy(&ppObject->front(), pData, (size_t)*pVectorSize);
        }
    };

    // LIST
    template<class ValueType>
    struct serialize_traits<std::list<ValueType> >
    {
        // pack by value 
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const std::list<ValueType> & object)
        {
            if (object.empty())
                pack_helper<ValueType>::PackEmptyArray(archive);
            else
            {
                int  * pTypeId = (int  *)archive.TailAllocate(sizeof (int));
                *pTypeId = rid_sequence;

                int  * pNumberOfDimensions = (int  *)archive.TailAllocate(sizeof (int));
                *pNumberOfDimensions = 1;

                rpc_size_t  * pSize = (rpc_size_t  *)archive.TailAllocate(sizeof (rpc_size_t));
                *pSize = object.size();

                for(std::list<ValueType>::const_iterator it = object.begin(); it != object.end(); ++it)
                    Pack(archive, *it);
            }
        }
        // pack by pointer
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const std::list<ValueType> * pObject)
        {
            serialize(archive, *pObject);
        }

        // unpack by ptr 
        template<class ArchiveType>
        static void deserialize(ArchiveType & archive, std::list<ValueType> * ppObject)
        {
            int  * pTypeId = (int  *)archive.FetchRange(sizeof (int));
            if (*pTypeId != rid_sequence)
                throw rpc::IncorrectTypeId(rid_sequence, *pTypeId);

            int  * pNumberOfDimensions = (int  *)archive.FetchRange(sizeof (int));
            if (*pNumberOfDimensions != 1)
                throw std::runtime_error("Array deserialization failed - incorrect dimension");

            rpc_size_t * pDim1 = (rpc_size_t  *)archive.FetchRange(sizeof (rpc_size_t));

            if (!*pDim1)
            {
                ppObject->clear();
                return;
            }

            for(rpc_size_t i = 0; i<*pDim1; ++i)
            {
                ppObject->push_back(ValueType());
                Unpack(archive, &ppObject->back());  
            }
        }
    };

    // MAP
    template<class KeyType, class ValueType>
    struct serialize_traits<std::map<KeyType,ValueType> >
    {
        typedef std::map<KeyType,ValueType> ContainerType;
        // pack by value 
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const ContainerType & object)
        {
            int  * pTypeId = (int  *)archive.TailAllocate(sizeof (int));
            *pTypeId = rid_map;

            rpc_size_t  * pSize = (rpc_size_t  *)archive.TailAllocate(sizeof (rpc_size_t));
            *pSize = object.size();

            if (!*pSize)
                return;

            for(ContainerType::const_iterator it = object.begin(); it != object.end(); ++it)
            {
                serialize_traits<ContainerType::key_type>::serialize(archive, it->first);
                serialize_traits<ContainerType::mapped_type>::serialize(archive, it->second);
            }
        }
        // pack by pointer
        template<class ArchiveType>
        static void serialize(ArchiveType & archive, const ContainerType * pObject)
        {
            serialize_traits<ObjectType>::serialize(archive, *pObject);
        }

        // unpack by ptr 
        template<class ArchiveType>
        static void deserialize(ArchiveType & archive, ContainerType * pObject)
        {
            int  * pTypeId = (int  *)archive.FetchRange(sizeof (int));
            if (*pTypeId != rid_map)
                throw rpc::IncorrectTypeId(rid_map, *pTypeId);

            rpc_size_t * pDim1 = (rpc_size_t  *)archive.FetchRange(sizeof (rpc_size_t));
            if (!*pDim1)
            {
                pObject->clear();
                return;
            }

            for(rpc_size_t i = 0; i<*pDim1; ++i)
            {
                std::pair<typename ContainerType::key_type, typename ContainerType::mapped_type> value;

                serialize_traits<ContainerType::key_type>::deserialize(archive, &value.first);  
                serialize_traits<ContainerType::mapped_type>::deserialize(archive, &value.second);  
                std::pair<ContainerType::iterator, bool> res = pObject->insert( value );
                if (!res.second)
                    throw std::runtime_error("Map deserialization failed - duplicated keys");
            }
        }
    };

}//rpc

#endif
