#ifndef TESTOBJECT_H__
#define TESTOBJECT_H__

#include <vector>

#include "RpcArchive.h"
#include "RpcSerialization.h"
#include "StlSerialization.h"
#include "RpcServer.h"
#include "RpcCall.h"
#include "RpcSkel.h"

const int rid_CustomException = RPC_USER_RID(1);

class CustomException: public std::runtime_error
{
    friend rpc::serialize_traits<CustomException>;

public:
    CustomException(const std::string& message, int customCode = 0)
        : message_(message), code_(customCode)
        , std::runtime_error(message)
    {
    }

    const char* what()
    {
        return message_.c_str();
    }

    int GetCode()
    {
        return code_;
    }

private:
    std::string message_;
    int code_;
};

namespace rpc
{
    template <>
    struct serialize_traits<CustomException>
    {
        template<class ArchiveType>
        static void serialize(ArchiveType& archive, const CustomException& object)
        {
            int* typeId = (int*)archive.TailAllocate(sizeof(int));
            *typeId = rid_CustomException;

            rpc::Pack(archive, object.message_);
            rpc::Pack(archive, object.code_);
        }

        template<class ArchiveType>
        static void deserialize(ArchiveType& archive, CustomException* object)
        {
            int* typeId = (int*)archive.FetchRange(sizeof(int));
            if (*typeId != rid_CustomException)
                throw rpc::IncorrectTypeId(rid_CustomException, *typeId);

            std::string message;
            rpc::Unpack(archive, &message);

            int code;
            rpc::Unpack(archive, &code);

            *object = CustomException(message, code);
        }
    };
}

//////////////////////////////////////////////////////////////////////////

class TestObject: public rpc::RpcSkelBase
{
    void TestMapMethod(rpc::Archive& inBuf, rpc::Archive& outBuf)
    {
        typedef std::map<std::string, int> TestMap;
        TestMap in_map;
        rpc::Unpack(inBuf, &in_map);

        TestMap::iterator it_m = in_map.find("test_item");
        if(it_m != in_map.end())    //insert new item
            in_map.insert(std::make_pair<std::string, int>("new_item", ++(it_m->second)));

        rpc::Pack(outBuf, in_map);
    }

    void TestValidityMethod(rpc::Archive& inBuf, rpc::Archive& outBuf)
    {
        int in_param = 0;
        rpc::Unpack(inBuf, &in_param);

        in_param++;

        rpc::Pack(outBuf, in_param);//some int param
    }

    void TestValidityWithoutParams(rpc::Archive& inBuf, rpc::Archive& outBuf)
    {
    }

    void TestExceptionMethod(rpc::Archive& inBuf, rpc::Archive& outBuf)
    {
        throw CustomException("Test exception", 42);
    }

    void TestStringMethod(rpc::Archive& inBuf, rpc::Archive& outBuf)
    {
        std::string in_str;
        std::string add_str("succeeded");

        rpc::Unpack(inBuf, &in_str);

        in_str += add_str;

        rpc::Pack(outBuf, in_str);
    }

    void TestVectorMethod(rpc::Archive& inBuf, rpc::Archive& outBuf)
    {
        std::vector<int> input_vr;
        rpc::Unpack(inBuf, &input_vr);

        for(std::vector<int>::iterator it = input_vr.begin(); it != input_vr.end(); ++it) ++(*it);

        rpc::Pack(outBuf, input_vr);
    }

public:
    TestObject(rpc::IRpcServer* rpcServer)
    {
        RPC_FUNCTION(TestObject, TestValidityMethod);
        RPC_FUNCTION(TestObject, TestExceptionMethod);
        RPC_FUNCTION(TestObject, TestStringMethod);
        RPC_FUNCTION(TestObject, TestVectorMethod);
        RPC_FUNCTION(TestObject, TestMapMethod);
        RPC_FUNCTION(TestObject, TestValidityWithoutParams);
        rpcServer->RegisterObject("test_object", this);
    }
};


#endif
